def Red(element):
  element.__color_scheme__ = {
    "bioc:stroke"            : "#831311",
    "bioc:fill"              : "#ffcdd2",
    "color:background-color" : "#ffcdd2",
    "color:border-color"     : "#831311"
  }
  return element

def Green(element):
  element.__color_scheme__ = {
    "bioc:stroke"            : "#205022",
    "bioc:fill"              : "#c8e6c9",
    "color:background-color" : "#c8e6c9",
    "color:border-color"     : "#205022"
  }
  return element

def Orange(element):
  element.__color_scheme__ = {
    "bioc:stroke"            : "#6b3c00",
    "bioc:fill"              : "#ffe0b2",
    "color:background-color" : "#ffe0b2",
    "color:border-color"     : "#6b3c00"
  }
  return element

def Blue(element):
  element.__color_scheme__ = {
    "bioc:stroke"            : "#0d4372",
    "bioc:fill"              : "#bbdefb",
    "color:background-color" : "#bbdefb",
    "color:border-color"     : "#0d4372"
  }
  return element
